from BayesNetCreacion import BayesNetCreacion, Node

#Burglar Alarm System example 
#Reference: https://www.youtube.com/watch?v=hEZjPZ-Ze0A&t=35s

#Creating nodes and network
burglary = Node('B')
earthquake = Node('E')
alarm = Node('A')
johncalls = Node('J')
marycalls = Node('M')

alarm.set_parents([str(burglary),str(earthquake)])
johncalls.set_parents([str(alarm)])
marycalls.set_parents([str(alarm)])

bnc = BayesNetCreacion()
bnc.add_node(burglary)
bnc.add_node(earthquake)
bnc.add_node(alarm)
bnc.add_node(johncalls)
bnc.add_node(marycalls)

bnc.add_prob({
    ('B', True): 0.001,
    ('B', False): 0.999,
    ('E', True): 0.002,
    ('E', False): 0.998,
    ('A', True, 'B', True, 'E', True): 0.95,
    ('A', True, 'B', False, 'E', True): 0.29,
    ('A', True, 'B', True, 'E', False): 0.94,
    ('A', True, 'B', False, 'E', False): 0.001,
    ('A', False, 'B', True, 'E', True): 0.05,
    ('A', False, 'B', False, 'E', True): 0.71,
    ('A', False, 'B', True, 'E', False): 0.06,
    ('A', False, 'B', False, 'E', False): 0.999,
    ('J', True, 'A', True): 0.9,
    ('J', True, 'A', False): 0.05,
    ('J', False, 'A', True): 0.1,
    ('J', False, 'A', False): 0.95,
    ('M', True, 'A', True): 0.7,
    ('M', True, 'A', False): 0.01,
    ('M', False, 'A', True): 0.3,
    ('M', False, 'A', False): 0.99,
})

# print(bnc.get_network())

#Functions
#Probabilistic Inference
evidence = {'J': True, 'M': True}
query = 'B'
result = bnc.probabilistic_inference(query, evidence)
print(result) #0.28417

#Compact


#Factores
print(bnc.get_factors()) #Outputs Conditional Probability Table

#Definition check


#Enumeration
evidence = {'B': True, 'E':False}
query = 'A'
result = bnc.pre_enum(query, evidence)
print(result) #{True: 0.94, False: 0.06}


# Base
class BayesNetCreacion():
    def __init__(self):
        """
        Class variables:
        :variables: List, Saves all the variables added to the network
        :parents: Dict, Saves all the parents of the variables added
        :cpt: Dict, contains alll the probabilities of the variables in the network
        """
        self.variables = ()
        self.parents = {}
        self.cpt = {}
    
    def add_node(self, node): #Adds node object to the network
        query = {str(node): node.get_parents()}
        self.parents.update(query)
        self.variables += tuple(query.keys())
    
    def add_prob(self, prob): #Adds a probability or cpt to the network
        self.cpt.update(prob)
    
    def get_network(self): #Returns a dict containing the network values
        return {
            'variables': self.variables,
            'parents': self.parents,
            'cpt' : self.cpt
        }

    def probabilistic_inference(self, query, evidence):
        """
        Calculate the probability of the query given the evidence inputted

        :param: query: the variable to calculate the prob for
        :param: evidence: the evidence, given a dictionary with their values
        :return: the probability of the query happening given the evidence
        """
        result = self.pre_enum(query, evidence)
        return result[True]

    def P(self, var, e):
        """
        Calculate the probability of a variable given its parents in the network, using the conditional probability table.

        :param var: the variable to calculate the probability for
        :param e: the evidence, a dictionary that maps variables to their values
        :return: the probability of the variable given the evidence
        """  
        key = (var, e[var])
        for p in self.parents[var]:
            key += tuple(p)
            key += tuple([e[p]])
        return self.cpt[key]

    #Additional services



    def get_factors(self): #Returns factors of the network (dict)
        return self.cpt

    def pre_enum(self, X, e): 
        QX = {}
        for xi in [True, False]:
            e[X] = xi
            QX[xi] = self.get_enum(self.variables, e)
        return self.normalize(QX)
    
    def get_enum(self, variables, e): #Returns numeration value(s)
        if not variables:
            return 1
        Y, rest = variables[0], variables[1:]
        if Y in e:
            return self.P(Y, e) * self.get_enum(rest, e)
        else:
            return sum(self.P(Y, self.extend(e, Y ,yi))* self.get_enum(rest, self.extend(e, Y, yi))
                        for yi in [True, False])

    def extend(self, e, var, val):
        e2 = dict(e)
        e2[var] = val
        return e2

    def normalize(self, QX): #Returns the normalized numeration value
        total = sum(QX.values())
        for key in QX:
            QX[key] /= total
        return QX


#Creation of each probabilistic node that will form the network
class Node():
    def __init__(self, name):
        """
        Class variables:
        :name: String, contains the name which identifies the node created.
        :parents: String list, contains the names of the parents of the node, is empty if node doesnt have parents.
        """
        self.name = name
        self.parents = [] 
    
    def __str__(self):
        return self.name

    def set_parents(self, parents: list):
        self.parents = parents
    
    def get_parents(self):
        return self.parents
    
    




from collections import defaultdict

class BayesianNetwork:
    def __init__(self, parents, cpt):
        """
        Initialize the Bayesian network with a dictionary of parents and a dictionary of conditional probability tables.

        :param variables: a list containing each variable from parents
        :param parents: a dictionary that maps each variable to its parent variables
        :param cpt: a dictionary that maps each variable-value pair and its parent variables-values to its probability
        """
        self.variables = list(parents.keys())
        self.parents = parents
        self.cpt = cpt

    def P(self, var, e):
        """
        Calculate the probability of a variable given its parents in the network, using the conditional probability table.

        :param var: the variable to calculate the probability for
        :param e: the evidence, a dictionary that maps variables to their values
        :return: the probability of the variable given the evidence
        """
        
        key = (var, e[var])
        for p in self.parents[var]:
            key += tuple(p)
            key += tuple([e[p]])
        return self.cpt[key]

def enumeration_ask(X, e, bn):
    QX = {}
    for xi in [True, False]:
        e[X] = xi
        QX[xi] = enumerate_all(bn.variables, e, bn)
    return normalize(QX)

def enumerate_all(variables, e, bn):
    if not variables:
        return 1.0
    Y, rest = variables[0], variables[1:]
    if Y in e:
        return bn.P(Y, e) * enumerate_all(rest, e, bn)
    else:
        return sum(bn.P(Y, extend(e, Y, yi)) * enumerate_all(rest, extend(e, Y, yi), bn)
                   for yi in [True, False])

def extend(e, var, val):
    e2 = dict(e)
    e2[var] = val
    return e2

def normalize(QX):
    total = sum(QX.values())
    for key in QX:
        QX[key] /= total
    return QX

# Define the Bayesian network
network = BayesianNetwork({
    'B': [],
    'E': [],
    'A': ['B', 'E'],
    'J': ['A'],
    'M': ['A']
}, {
    ('B', True): 0.001,
    ('B', False): 0.999,
    ('E', True): 0.002,
    ('E', False): 0.998,
    ('A', True, 'B', True, 'E', True): 0.95,
    ('A', True, 'B', False, 'E', True): 0.29,
    ('A', True, 'B', True, 'E', False): 0.94,
    ('A', True, 'B', False, 'E', False): 0.001,
    ('A', False, 'B', True, 'E', True): 0.05,
    ('A', False, 'B', False, 'E', True): 0.71,
    ('A', False, 'B', True, 'E', False): 0.06,
    ('A', False, 'B', False, 'E', False): 0.999,
    ('J', True, 'A', True): 0.9,
    ('J', True, 'A', False): 0.05,
    ('J', False, 'A', True): 0.1,
    ('J', False, 'A', False): 0.95,
    ('M', True, 'A', True): 0.7,
    ('M', True, 'A', False): 0.01,
    ('M', False, 'A', True): 0.3,
    ('M', False, 'A', False): 0.99,

})

# Perform probabilistic inference
evidence = {'J': True, 'M': True}
query = 'B'
result = enumeration_ask(query, evidence, network)
print(result[True])




from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()

VERSION = '1.1.1'
DESCRIPTION = 'Library capable of creating Bayesian Networks and making probabilistic inference over them, as well additional functions'
LONG_DESCRIPTION = 'This library has the objective of building Bayesian networks and making probabilistic inference over them. Also, adding some other additional features that could serve the developers that make use of this library. This library has zero dependencies to assure it is futureproof, easier to debug, to contribute to and use. For the most part this library works over classes like BayesNetCreacion and Node, this was chosen so that in a way it could facilitate the usage of the OOP paradigm.'

# Setting up
setup(
    name="BayesNetCreacion",
    version=VERSION,
    author="Andres de la Roca",
    author_email="<dela20332@uvg.edu.gt>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    keywords=['python', 'bayes', 'bayesian network', 'zero dependencies', 'bayesian inference', 'python3', 'open source'],
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Microsoft :: Windows",
    ]
)